@@ -30,7 +30,7 @@ except ImportError:
 
 Config = namedtuple('Config', ('layers', 'codewords', 'cw_bits', 'bits', 'digits', 'text', 'bytes'))
 
-table = {
+configs = {
     (15, True): Config(layers=1, codewords=17, cw_bits=6, bits=102, digits=13, text=12, bytes=6),
     (19, False): Config(layers=1, codewords=21, cw_bits=6, bits=126, digits=18, text=15, bytes=8),
     (19, True): Config(layers=2, codewords=40, cw_bits=6, bits=240, digits=40, text=33, bytes=19),
@@ -90,7 +90,7 @@ code_chars = {
     Mode.DIGIT: [Shift.PUNCT] + list(b' 0123456789,.') + [Latch.UPPER, Shift.UPPER],
 }
 
-punct_2_chars = [pc for pc in code_chars[Mode.PUNCT] if isinstance(pc, bytes) and len(pc) == 2]
+punct_2_chars = [pc for pc in code_chars[Mode.PUNCT] if isinstance(pc, bytes)]
 
 E = 99999  # some big number
 
@@ -228,7 +228,9 @@ def find_optimal_sequence(data):
                     elif cur_seq[x]:
                         # if changing from punct or digit mode - use U/L as intermediate mode
                         # TODO: update for digit
-                        if x in (Mode.PUNCT, Mode.DIGIT) and y != Mode.UPPER:
+                        if x == Mode.DIGIT and y == Mode.PUNCT:
+                            cur_seq[y] += [Misc.RESUME, Latch.UPPER, Latch.MIXED, Latch.PUNCT]
+                        elif x in (Mode.PUNCT, Mode.DIGIT) and y != Mode.UPPER:
                             cur_seq[y] += [Misc.RESUME, Latch.UPPER, Latch[y.name]]
                         elif x in (Mode.UPPER, Mode.LOWER) and y == Mode.PUNCT:
                             cur_seq[y] += [Latch.MIXED, Latch[y.name]]
@@ -313,7 +315,7 @@ def find_optimal_sequence(data):
         prev_c = c
     # sort in ascending order and get shortest sequence
     result_seq = []
-    sorted_cur_len = sorted(cur_len, key=cur_len.get)
+    sorted_cur_len = sorted(cur_len, key=cur_len.__getitem__)
     if sorted_cur_len:
         min_length = sorted_cur_len[0]
         result_seq = cur_seq[min_length]
@@ -358,39 +360,23 @@ def optimal_sequence_to_bits(optimal_sequence):
     :return: string with bits
     """
     out_bits = ''
-    mode = Mode.UPPER
-    prev_mode = Mode.UPPER
+    mode = prev_mode = Mode.UPPER
     shift = False
-    binary = False
-    binary_seq_len = 0
-    binary_index = 0
     sequence = optimal_sequence[:]
-    while True:
-        if not sequence:
-            break
+    while sequence:
         # read one item from sequence
         ch = sequence.pop(0)
-        if binary:
-            out_bits += bin(ch)[2:].zfill(char_size.get(mode))
-            binary_index += 1
-            # resume previous mode at the end of the binary sequence
-            if binary_index >= binary_seq_len:
-                mode = prev_mode
-                binary = False
-            continue
-        index = code_chars.get(mode).index(ch)
-        out_bits += bin(index)[2:].zfill(char_size.get(mode))
+        index = code_chars[mode].index(ch)
+        out_bits += bin(index)[2:].zfill(char_size[mode])
         # resume previous mode for shift
         if shift:
             mode = prev_mode
             shift = False
         # get mode from sequence character
-        if ch in Latch or ch in Shift:
+        if ch in Latch:
             mode = ch.value
-        if ch in Shift:
-            shift = True
         # handle FLG(n)
-        if ch == Misc.FLG:
+        elif ch == Misc.FLG:
             if not sequence:
                 raise Exception('Expected FLG(n) value')
             flg_n = sequence.pop(0)
@@ -400,19 +386,19 @@ def optimal_sequence_to_bits(optimal_sequence):
                 raise Exception('FLG(7) is reserved and currently illegal')
 
             out_bits += bin(flg_n)[2:].zfill(3)
-            if flg_n > 1:
+            if flg_n >= 1:
                 # ECI
                 if not sequence:
                     raise Exception('Expected FLG({}) to be followed by ECI code'.format(flg_n))
                 eci_code = sequence.pop(0)
                 if not isinstance(eci_code, numbers.Number) or not 0 <= eci_code < (10**flg_n):
                     raise Exception('Expected FLG({}) ECI code to be a number from 0 to {}'.format(flg_n, (10**flg_n) - 1))
-                out_digits = str(eci_code).zfill(flg_n).encode()
+                out_digits = b'%.*d' % (flg_n, eci_code)
                 for ch in out_digits:
                     index = code_chars[Mode.DIGIT].index(ch)
                     out_bits += bin(index)[2:].zfill(char_size[Mode.DIGIT])
-        # handle binary mode
-        if mode == Mode.BINARY:
+        # handle binary run
+        elif ch == Shift.BINARY:
             if not sequence:
                 raise Exception('Expected binary sequence length')
             # followed by a 5 bit length
@@ -420,19 +406,20 @@ def optimal_sequence_to_bits(optimal_sequence):
             if not isinstance(seq_len, numbers.Number):
                 raise Exception('Binary sequence length must be a number')
             out_bits += bin(seq_len)[2:].zfill(5)
-            binary_seq_len = seq_len
             # if length is zero - 11 additional length bits are used for length
-            if not binary_seq_len:
+            if not seq_len:
                 seq_len = sequence.pop(0)
                 if not isinstance(seq_len, numbers.Number):
                     raise Exception('Binary sequence length must be a number')
                 out_bits += bin(seq_len)[2:].zfill(11)
-                binary_seq_len = seq_len + 31
-            binary = True
-            binary_index = 0
-        # update previous mode
-        if not shift:
-            prev_mode = mode
+                seq_len += 31
+            for binary_index in range(seq_len):
+                ch = sequence.pop(0)
+                out_bits += bin(ch)[2:].zfill(char_size[Mode.BINARY])
+        # handle other shift
+        elif ch in Shift:
+            mode, prev_mode = ch.value, mode
+            shift = True
     return out_bits
 
 
@@ -469,36 +456,34 @@ def get_data_codewords(bits, codeword_size):
 
 
 def get_config_from_table(size, compact):
-    """ Get config from table with given size and compactness flag
+    """ Get config with given size and compactness flag
 
     :param size: matrix size
     :param compact: compactness flag
     :return: dict with config
     """
-    config = table.get((size, compact))
-    if not config:
-        raise Exception('Failed to find config with size and compactness flag')
-    return config
+    try:
+        return configs[(size, compact)]
+    except KeyError:
+        raise NotImplementedError('Failed to find config with size and compactness flag')
 
-def find_suitable_matrix_size(data):
+def find_suitable_matrix_size(data, ec_percent=23):
     """ Find suitable matrix size
     Raise an exception if suitable size is not found
 
     :param data: string or bytes to encode
+    :param ec_percent: percentage of symbol capacity for error correction (default 23%)
     :return: (size, compact) tuple
     """
     optimal_sequence = find_optimal_sequence(data)
     out_bits = optimal_sequence_to_bits(optimal_sequence)
-    for (size, compact) in sorted(table.keys()):
+    for (size, compact) in configs.keys():
         config = get_config_from_table(size, compact)
         bits = config.bits
-        # error correction percent
-        ec_percent = 23  # recommended: 23% of symbol capacity plus 3 codewords
         # calculate minimum required number of bits
-        required_bits_count = int(math.ceil(len(out_bits) * 100.0 / (
-            100 - ec_percent) + 3 * 100.0 / (100 - ec_percent)))
+        required_bits_count = int(math.ceil((len(out_bits) + 3) * 100.0 / (100 - ec_percent)))
         if required_bits_count < bits:
-            return size, compact
+            return size, compact, optimal_sequence
     raise Exception('Data too big to fit in one Aztec code!')
 
 class AztecCode(object):
@@ -506,7 +491,7 @@ class AztecCode(object):
     Aztec code generator
     """
 
-    def __init__(self, data, size=None, compact=None):
+    def __init__(self, data, size=None, compact=None, ec_percent=23):
         """ Create Aztec code with given data.
         If size and compact parameters are None (by default), an
         optimal size and compactness calculated based on the data.
@@ -514,8 +499,11 @@ class AztecCode(object):
         :param data: string or bytes to encode
         :param size: size of matrix
         :param compact: compactness flag
+        :param ec_percent: percentage of symbol capacity for error correction (default 23%)
         """
         self.data = data
+        self.sequence = None
+        self.ec_percent = ec_percent
         if size is not None and compact is not None:
             if (size, compact) in table:
                 self.size, self.compact = size, compact
@@ -523,7 +511,7 @@ class AztecCode(object):
                 raise Exception(
                     'Given size and compact values (%s, %s) are not found in sizes table!' % (size, compact))
         else:
-            self.size, self.compact = find_suitable_matrix_size(self.data)
+            self.size, self.compact, self.sequence = find_suitable_matrix_size(self.data, ec_percent)
         self.__create_matrix()
         self.__encode_data()
 
@@ -531,19 +519,21 @@ class AztecCode(object):
         """ Create Aztec code matrix with given size """
         self.matrix = [array.array('B', (0 for jj in range(self.size))) for ii in range(self.size)]
 
-    def save(self, filename, module_size=1, border=0):
+    def save(self, filename, module_size=1, border=0, format=None):
         """ Save matrix to image file
 
-        :param filename: output image filename.
+        :param filename: output image filename (or file object, with format).
         :param module_size: barcode module size in pixels.
         :param border: barcode border size in modules.
+        :param format: Pillow image format, such as 'PNG'
         """
-        self.image(module_size, border).save(filename)
+        self.image(module_size, border).save(filename, format=format)
 
     def image(self, module_size=1, border=0):
         """ Create PIL image
 
         :param module_size: barcode module size in pixels.
+        :param border: barcode border size in modules
         """
         if ImageDraw is None:
             exc = missing_pil[0](missing_pil[1])
@@ -559,10 +549,25 @@ class AztecCode(object):
                     fill=not self.matrix[y][x])
         return image
 
-    def print_out(self):
-        """ Print out Aztec code matrix """
-        for line in self.matrix:
-            print(''.join(('#' if x else ' ') for x in line))
+    def print_out(self, border=0):
+        """ Print out Aztec code matrix using ASCII output """
+        if not fancy:
+            print('\n'.join(' '*(2*border + self.size) for ii in range(border)))
+            for line in self.matrix:
+                print(' '*border + ''.join(('#' if x else ' ') for x in line) + ' '*border)
+            print('\n'.join(' '*(2*border + self.size) for ii in range(border)))
+
+    def print_fancy(self, border=0):
+        """ Print out Aztec code matrix using Unicode box-drawing characters and ANSI colorization """
+        for y in range(-border, self.size+border, 2):
+            last_half_row = (y==self.size + border - 1)
+            ul = '\x1b[40;37;1m' + ('\u2580' if last_half_row else '\u2588')*border
+            for x in range(0, self.size):
+                a = self.matrix[y][x] if 0 <= y < self.size else None
+                b = self.matrix[y+1][x] if -1 <= y < self.size-1 else last_half_row
+                ul += ' ' if a and b else '\u2584' if a else '\u2580' if b else '\u2588'
+            ul += ('\u2580' if last_half_row else '\u2588')*border + '\x1b[0m'
+            print(ul)
 
     def __add_finder_pattern(self):
         """ Add bulls-eye finder pattern """
@@ -681,8 +686,9 @@ class AztecCode(object):
         :param data: data to encode
         :return: number of data codewords
         """
-        optimal_sequence = find_optimal_sequence(data)
-        out_bits = optimal_sequence_to_bits(optimal_sequence)
+        if not self.sequence:
+            self.sequence = find_optimal_sequence(data)
+        out_bits = optimal_sequence_to_bits(self.sequence)
         config = get_config_from_table(self.size, self.compact)
         layers_count = config.layers
         cw_count = config.codewords
@@ -798,16 +804,18 @@ class AztecCode(object):
         self.__add_mode_info(data_cw_count)
 
 
-def main():
-    data = 'Aztec Code 2D :)'
+def main(argv):
+    if len(argv) not in (2, 3):
+        print("usage: {} STRING_TO_ENCODE [IMAGE_FILE]".format(argv[0]))
+        print("  Generate a 2D Aztec barcode and print it, or save to a file.")
+    data = argv[1]
     aztec_code = AztecCode(data)
-    aztec_code.print_out()
-    if ImageDraw is None:
-        print('PIL is not installed, cannot generate PNG')
-    else:
-        aztec_code.save('aztec_code.png', 4)
     print('Aztec Code info: {0}x{0} {1}'.format(aztec_code.size, '(compact)' if aztec_code.compact else ''))
+    if len(sys.argv) == 3:
+        aztec_code.save(argv[2], module_size=5)
+    else:
+        aztec_code.print_fancy(border=2)
 
 
 if __name__ == '__main__':
-    main()
+    main(sys.argv)
